/*
 *----------------------------------------------------------------------------
 *	LSI Corporation
 *	1621 Barber Lane
 *	Milpitas, California 95035
 *----------------------------------------------------------------------------
 * Copyright  2004-2006, LSI Corporation All Rights Reserved.
 * 
 * LSI's source code is an unpublished work and the use of copyright
 * notice does not imply otherwise. This source code contains confidential,
 * trade secret material of LSI Corporation. Any attempt or
 * participation in deciphering, decoding, reverse engineering or
 * in any way altering the source code is strictly prohibited, unless the
 * prior written consent of LSI Corporation.
 *---------------------------------------------------------------------------- 
 */

/** 
 *@file linux_osl.c
 *
 *@brief
 * This file contains the Linux OSL implementation.
 *
 * Linux Operating System related implementation is present in this file. This
 * file will contain driver entry points and the call back functions to the
 * SCSI Mid Layer. Linux Operating system Layer functions are also implemented
 * in this file 
 * 
 *@bug
 *	None
 *
 *@warning
 *	None
 *
 */

#include "linux_osl.h"
#include "osl_char.h"
#include "../../cmn_defs.h"
#include "../../oss/oss_iface.h"
#include "../../swr5_pciids.h"

#define LINOSL_MAX_IDS					20

typedef struct _adapter_t adapter_t;
extern int			adp_count_g ;
extern adapter_t	*adps_g[];

MODULE_AUTHOR		("LSI Corporation");

#ifdef MODULE_VERSION
MODULE_VERSION(LSIRAID_VERSION_STRING);
#endif /*MODULE_VERSION*/

#ifdef LSI_RAID5_SUPPORT
 MODULE_DESCRIPTION	("LSI Software RAID5 Driver");
#else
 MODULE_DESCRIPTION	("LSI Software RAID Driver");
#endif

MODULE_LICENSE		("LSI Proprietary");

uint32_t			dbglvl_g = DRIVER_DBG_LEVEL;
struct fasync_struct *lsraid_async_queue;
//static DECLARE_MUTEX(lsraid_async_queue_mutex);


#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,0)
 module_param(dbglvl_g, uint, DRIVER_DBG_LEVEL);
#else 
#ifdef MODULE
 MODULE_PARM			(dbglvl_g, "i");
#endif
#endif

MODULE_PARM_DESC	(dbglvl_g, "driver-wide debug flag");

/* supported controller ids */
static struct pci_device_id lsraid_pci_tbl_g[] = SWR5_PCI_IDS;
MODULE_DEVICE_TABLE(pci, lsraid_pci_tbl_g);

extern uint8_t num_pci_dev_table_entries_g;

#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,0)
static struct scsi_host_template	driver_template = LSI_RAID_TEMPLATE;
static struct scsi_host_template	*lsraid_template_gp	= &driver_template;
#else
static Scsi_Host_Template	driver_template = LSI_RAID_TEMPLATE;
static Scsi_Host_Template	*lsraid_template_gp	= &driver_template;
#endif

/* Function Prototypes */
static LSI_RAID_IRQRETURN_T lsraid_isr(int irq, void *devp, struct pt_regs *regs);

/* helper functions */
static int __devinit lsraid_probe_one(struct pci_dev*, 
		const struct pci_device_id *);
static void lsraid_detach_one(struct pci_dev *dev);
void __lsraid_shutdown(void *adp);
static void	extract_pciconfig (struct pci_dev*, pvt_data_t *);
void remove_mem_allocated_for_memmgr(linux_adp_t	*lxadp);
boolean_t allocate_static_mem_and_add_to_memmgr(linux_adp_t *lxadp);
static uint32_t	scmd_to_rclpkt (struct scsi_cmnd*, rcl_packet_public_t*, struct pci_dev *);
static void_t osl_destroy_all_timers(linux_adp_t *lxadp);
void	osl_timeout (unsigned long);

void_t osl_cmd_pkt_success_handler(pvt_data_t* os_cxt, pvt_data_t* rclp_ptr);

rcl_packet_public_t *rcl_alloc_rcl_packet(void *adp);
boolean_t	osl_cmn_abort_cmd(void *adp, rcl_packet_public_t *rcl_pkt);
boolean_t	osl_cmn_reset(void *adp, uint8_t reset_lvl, device_id_t *device_id);
uint32_t megasr_lib_isr(void_t *adp);
void_t osl_cmn_dump_busy_queue(void *adp, uint32_t rcl_error_status);
uint32_t megasr_lib_get_adp_size(void);
void megasr_lib_set_lxadp(void *adp, void *lxadp);
extern uint32_t megasr_init_adp(pvt_data_t *adpv, uint8_t bus_number, uint8_t dev_no, uint8_t fn_no);

uint32_t	rcl_scan_devices(void *adp);
//void osl_log_adp_pci_info(void *adp);
uint32_t	rcl_read_and_populate_cfg(void *adp);
void_t rcl_unload(void *adp);
void_t		rcl_shutdown(void *adp);
void *lin_get_adp_pci_config_space_ptr_and_size(void *adpv, uint32_t *cfg_size);
void lin_set_adp_pci_config(uint32_t base_addr,int i, void *adp);
void_t rcl_free_rcl_packet(pvt_data_t *adp, rcl_packet_public_t *rclp);
pvt_data_t * lin_get_linux_adp(void *data);
uint32_t	rcl_init_adapter(pvt_data_t *adp);
boolean_t mem_mgr_add_sgl_mem(pvt_data_t *adapter, cmn_sgl_t *sgl, pvt_data_t
*cxt);
void mem_mgr_fini_mem_mgr(pvt_data_t *adapter);
uint32_t rcl_is_r5rw_or_noniocmd(pvt_data_t *adp, rcl_packet_public_t *rclp);
void lin_dec_cmd_in_rcl_cnt(void *adp);
void megasr_lib_get_ven_dev_id_from_pci_dev_tbl(uint32_t dev_ind,
									uint16_t *venid, uint16_t *devid);

void lin_set_rclp_ld_ptr(pvt_data_t *adpv, pvt_data_t *rclpv, device_id_t *id);

void megasr_lib_init_adp(pvt_data_t *adpv, short *cmd_per_lun, short unsigned int *max_sectors, uint32_t adp_ind);
void_t		osl_cmn_process_rclp(pvt_data_t *adp, pvt_data_t *rclp);
uint8_t		osl_cmn_fill_sense_info_from_rclp(pvt_data_t *rclp, void_t *sense_data, uint8_t sense_len);
void_t		rcl_authenticate_ibutton(pvt_data_t *adp);
void_t		osl_cmn_flush_schedule_queue(pvt_data_t *adp);

void_t mem_mgr_init_mem_mgr(void *adapter, uint32_t node_size);
uint8_t mem_mgr_add_to_mem_pool(void_t *adp, uint8_t *ptr, uint8_t *phy_ptr, uint32_t tot_len, mem_pool_type_e pool_type);
uint32_t mem_mgr_get_desc_size_for_sgl_mem(pvt_data_t *adapter, uint32_t sgl_size);

/*
 * When we are asked to load on a controller that we didn't know about at compile time,
 * we use the megasr_adapter_class value to know the type of the controller. Users can
 * change this when they compile shim package
 */
megasr_adapter_class_t megasr_adapter_class = MEGASR_DYNAMIC_PCIIDS_DEFAULT_ADP_CLASS;

#if LINUX_VERSION_CODE < KERNEL_VERSION(2,6,0)

static int lsraid_reboot_notify (struct notifier_block *this, 
		unsigned long code, void *unused);

static struct notifier_block lsraid_reboot_notifier = {
	notifier_call : lsraid_reboot_notify
};

static struct proc_dir_entry	*megasas_proc_root;
static struct proc_dir_entry	*megasas_proc_hba_map;
static struct proc_dir_entry	*megasas_proc_version;
static struct proc_dir_entry	*megasas_proc_release_date;

static int megasas_proc_show_version(char *page, char **start, off_t offset,
		int count, int *eof, void *data)
{	
	*eof = 1;
	return (snprintf(page, strlen(LSIRAID_VERSION_STRING)+2, "%s\n",
				LSIRAID_VERSION_STRING));
}

static int megasas_proc_show_release_date(char *page, char **start, off_t
		offset, int count, int *eof, void *data)
{
	*eof = 1;
	return (snprintf(page, strlen(LSIRAID_DRIVER_RELEASE_DATE)+2, "%s\n",
				LSIRAID_DRIVER_RELEASE_DATE));
}

static int megasas_proc_show_bus_devfn(char *page, char **start, off_t offset,
		int count, int *eof, void *data)
{
	int		i;
	uint32_t	bus_devfn = 0;
	linux_adp_t *lxadp = (linux_adp_t *)data;

	bus_devfn = ((lxadp->pdev->bus->number) << 16 |
		(PCI_SLOT(lxadp->pdev->devfn)) << 8 |
		(PCI_FUNC(lxadp->pdev->devfn))) & 0xFFFFFF;

	*eof = 1;
	return (sprintf(page, "%d", bus_devfn));
}

static void megasas_create_proc_entry(void *lxp)
{
	linux_adp_t			*lxadp = (linux_adp_t *)lxp;
	unsigned char		string[8] = { 0 };

	sprintf(string, "%d", lxadp->host->host_no);
	lxadp->hba_proc_dir = proc_mkdir(string, megasas_proc_hba_map);

	if (!lxadp->hba_proc_dir) {
		con_log(CL_ERR, ("megaswr[osl]:   proc creation failed\n"));
		return;
	}

	create_proc_read_entry("bus_devfn", S_IRUSR | S_IFREG,
			lxadp->hba_proc_dir, megasas_proc_show_bus_devfn, lxadp );
}

static void megasas_remove_proc_entry(pvt_data_t *lxp)
{
	linux_adp_t		*lxadp = (linux_adp_t *)lxp;
	unsigned char		string[8] = { 0 };

	remove_proc_entry("bus_devfn", lxadp->hba_proc_dir);

	sprintf(string, "%d", lxadp->host->host_no);
	remove_proc_entry(string, megasas_proc_hba_map);
}

static void megasr_setup_proc(void)
{
	megasas_proc_root = proc_mkdir( LSIRAID_DRIVER_SIGNATURE, proc_scsi );

	if (megasas_proc_root) {
		megasas_proc_version = create_proc_read_entry( "version",
				S_IRUSR|S_IFREG, megasas_proc_root, megasas_proc_show_version,
				NULL);

		if (!megasas_proc_version){ 
			con_log(CL_ERR, ("megaswr[osl]:  cann't create_read_version p_version=%p\n", megasas_proc_version));
		}

		megasas_proc_release_date = create_proc_read_entry( "release_date",
				S_IRUSR | S_IFREG, megasas_proc_root,
				megasas_proc_show_release_date, NULL);

		megasas_proc_hba_map = proc_mkdir("hba_map",megasas_proc_root);
	}
}

#endif

/**
 * @brief
 * Private implementation for pci_alloc_consistent.
 * We do not want the GFP_ATOMIC flag to be set if caller can afford to sleep
 *
 * @param hwdev		: pointer to our device object
 * @param size		: size (in bytes) of memory required
 * @param linux_dma	: internal dma object tracker
 * @param wait		: are we allowed to sleep
 **/
uint32_t megasr_dma_alloc(struct pci_dev *hwdev, size_t size, linux_dma_mem_list_t *linux_dma, uint8_t wait)
{
	void	*rc;

#if LINUX_VERSION_CODE < KERNEL_VERSION(2,6,0)
	if (wait == LINUX_DMA_MEM_GFP) {
		rc = (void *)__get_free_pages(GFP_KERNEL, get_order(size));

		if (rc != NULL) {
			memset(rc, 0, size);
			linux_dma->vir		= (uint8_t *)rc;
			linux_dma->dma_addr	= virt_to_bus(rc);
			linux_dma->len		= (uint32_t)size;
			linux_dma->flag		= LINUX_DMA_MEM_GFP;
			return LSI_TRUE;
		} /* else fall through */
	} /* else fall through */
#endif

	// try to use the generic API first, to get the memory.
	// If that fail's, force use of the GFP_KERNEL flag
	rc = pci_alloc_consistent(hwdev, size, &linux_dma->dma_addr);

	if (rc) {
		linux_dma->vir	= (uint8_t *)rc;
		linux_dma->len	= (uint32_t)size;
		linux_dma->flag	= LINUX_DMA_MEM_PCI_CONSISTENT;
		return LSI_TRUE;
	}
	else if (wait == LINUX_DMA_NOWAIT) {
		// Cannot make the next calls if caller cannot sleep
		return LSI_FALSE;
	} /* else fall through */

	// Generic API failed to allocate memory, switch to alternate
#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,0)
	rc = dma_alloc_coherent(hwdev == NULL ? NULL : &hwdev->dev, size, &linux_dma->dma_addr, GFP_KERNEL|__GFP_REPEAT);
	if (rc != NULL) {
		linux_dma->vir		= (uint8_t *)rc;
		linux_dma->len		= (uint32_t)size;
		linux_dma->flag		= LINUX_DMA_MEM_PCI_COHERENT;
		return LSI_TRUE;
	}
#endif	
	return LSI_FALSE;
}

void megasr_dma_free(struct pci_dev *hwdev, linux_dma_mem_list_t *linux_dma)
{
	switch (linux_dma->flag) {

	case LINUX_DMA_MEM_PCI_CONSISTENT:
		pci_free_consistent(hwdev, (size_t)linux_dma->len, linux_dma->vir, linux_dma->dma_addr);
		break;

#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,0)
	case LINUX_DMA_MEM_PCI_COHERENT:
		dma_free_coherent(&hwdev->dev, (size_t)linux_dma->len, linux_dma->vir, linux_dma->dma_addr);
		break;
#else
	case LINUX_DMA_MEM_GFP:
		free_pages((unsigned long)linux_dma->vir, get_order(linux_dma->len));
		break;
#endif

	default:
		con_log(CL_ERR, ("megasr[%s]: invalid dma memory type for de-allocation\n", __FUNCTION__));
	}

	return;
}

#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,0)
#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,16)
static void lsraid_shutdown(struct pci_dev *dev);
#else
static void lsraid_shutdown(struct device *dev);
#endif
#endif

#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,18)
int megasr_resume( struct pci_dev *pdev)
{
	return 0;
}

int megasr_suspend(struct pci_dev *pdev, pm_message_t state)
{
	return -1;	//We don't support power management in Linux yet.
}
#endif

struct pci_driver lsraid_pci_driver_g = {
	.name		= LSIRAID_DRIVER_SIGNATURE,
	.id_table	= lsraid_pci_tbl_g,
	.probe		= lsraid_probe_one,
	.remove		= __devexit_p(lsraid_detach_one),
#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,0)
#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,16)
	.shutdown	= lsraid_shutdown,
#else
	.driver		= {
		.shutdown = lsraid_shutdown,
	}
#endif

#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,18) // Mainly for RHEL5 Gold
	.resume		= megasr_resume,
	.suspend	= megasr_suspend,
#endif
#endif
};

/**
 *@brief
 * Driver module entry point for 2.6 kernel.
 * This is called while driver is loading. This function prepares the PCI ID table
 * of the supported controllers and register driver module with the kernel.
 * This function also loads the character driver to support IOCTLs.
 *
 *@param		None
 *@return 		zero if function succeeds
 *@return 		no-zero if function fails
 */
static int __init lsraid_init(void)
{
	int	error;
	
	/* Announce the Driver Version */
	LSI_RAID_DISPLAY_DRIVER_VERSION;

#if LINUX_VERSION_CODE < KERNEL_VERSION(2,6,0)
	megasr_setup_proc();

	/* Register the Shutdown Notification hook in kernel */
	if(register_reboot_notifier(&lsraid_reboot_notifier)) {
		con_log(CL_ERR, ("megaswr[osl]:  reboot_notifier not registered!\n"));
	}
#endif

	error = pci_module_init(&lsraid_pci_driver_g);

	if (error) {

#if LINUX_VERSION_CODE < KERNEL_VERSION(2,6,0)
		unregister_reboot_notifier(&lsraid_reboot_notifier);

		if (megasas_proc_root) {
			remove_proc_entry("version", megasas_proc_root);
			remove_proc_entry("release_date", megasas_proc_root);
			remove_proc_entry("hba_map", megasas_proc_root);
			remove_proc_entry(LSIRAID_DRIVER_SIGNATURE, proc_scsi);
		}
#endif
		con_log(CL_ERR, ("megaswr[osl]: driver initialization error...\n"));
		return error;
	}
	
	lsraid_register_char_driver();

	return 0;
}

/**
 *@brief
 * Driver module exit point for 2.6 kernel.
 * This is called by the kernels while unloading the driver. This function
 * will unregister character driver loaded for supporting IOCTLs and
 * unregister the driver module itself from the kernel.
 * 
 *@param	None
 *@return 	None	
 */
static void __exit lsraid_exit(void)
{
	lsraid_unregister_char_driver();
	pci_unregister_driver(&lsraid_pci_driver_g);
}

/**
 *@brief
 * Driver shut down routine in for 2.6 kernel.
 * This function make use of generic shut down routine for both 2.6 and 2.4
 * kernels. This is called by the kernel when the system reboots and it is
 * done to let the lower level drivers to take all house keeping actions like
 * flushing the cache.
 * 
 *@param	dev		pointer to device structure
 *@return	None
 */
#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,0)
#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,16)

static void lsraid_shutdown(struct pci_dev *dev)
{
	struct Scsi_Host *host = pci_get_drvdata(dev);

	__lsraid_shutdown(((linux_adp_t *)host->hostdata[0])->adp);
}

#else /*LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,16)*/

static void lsraid_shutdown(struct device *dev)
{
	struct Scsi_Host *host = pci_get_drvdata(to_pci_dev(dev));

	__lsraid_shutdown(((linux_adp_t *)host->hostdata[0])->adp);
}

#endif /*LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,16)*/
#endif

#if LINUX_VERSION_CODE < KERNEL_VERSION(2,6,0)

/* Function Prototypes used only by 2.4 kernel */
/**
 *@brief
 * LSI MegaSWR release adapter
 * This function is used to unload MegaSWR driver for an adapter. Release
 * funtion will be called for each adapter present in the system. This is
 * called when the driver unloaded by using a rmmod command. This function
 * should cal reboot notifier of the driver when driver for all the
 * controllers are removed and it should unregister the character driver too.
 * 
 *@param	host	Pointer to scsi midlayer host structure
 *@return	0		If Success 
 */
int lsraid_release(struct Scsi_Host* host)
{
	linux_adp_t	*lxadp = (linux_adp_t *)host->hostdata[0];

	lsraid_detach_one(lxadp->pdev);
	
	if (!adp_count_g) {
		unregister_reboot_notifier(&lsraid_reboot_notifier);
		lsraid_unregister_char_driver();
	}

	return 0;
}

/**
 *@brief
 * This routine will be called when the user has done a forced shutdown on the
 * system. Flush the disks cache. This is called by the kernel when the system 
 * reboots and it is done to let the lower level drivers to take all house
 * keeping actions like flushing the cache.
 *
 *@param	this	Notifier Block structure pointer	
 *@param	code	shutdown code
 *@param	nused	
 *@return	NOTIFY_DONE after finishing 
 */
static int lsraid_reboot_notify (struct notifier_block *this, 
		unsigned long code, void *unused)
{
	int		i;

	con_dbg(CL_ANN, ("megaswr[osl]:  reboot notifier called\n"));

	for (i = 0; i < adp_count_g; i++) {
		__lsraid_shutdown(adps_g[i]);
	}

	return NOTIFY_DONE;
}
#endif /* LINUX_VERSION_CODE < KERNEL_VERSION(2,6,0) */

/**
 *@brief
 * LSI megasr info 
 * This function provide information string about the driver to the SCSI mid
 * layer.
 *
 *@param	host	megasr host structure pointer
 *@return	Driver information string
 *@remarks
 * This is a driver entry function
 */
const char *lsraid_info(struct Scsi_Host *host)
{
	return LSIRAID_DRIVER_NAME;
}


#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,0)
/**
 * @brief
 * megasr_linux_build_io
 * The "sleepy" build IO routine. This routine traverses the command in it's dedicated
 * queue and try to obtain the kernel virtual addresses for each IO packet found. We move
 * all such packets to our private list so that they can be processed at their own pace.
 * This is a tasklet, Linux guarantees, we are not running on more than one CPU at the
 * same time.
 **/
static void megasr_linux_build_io(void *devp)
{
	linux_adp_t				*linux_adp = (linux_adp_t *) devp;
	rcl_packet_public_t			*rclp;
	rcl_packet_public_t			*rclp_tmp;
	struct scsi_cmnd		*scp;
#ifdef LSI_RAID5_SUPPORT
	big_cmn_sgl_t				*sgl;
#else
	cmn_sgl_t				*sgl;
#endif
	unsigned long			flags;
	struct scatterlist		*mlsgl;
	struct megasr_clist	clist;
	int						i, j;

	megasr_clist_setup(&clist);

	spin_lock_irqsave(linux_adp->host_lock, flags);

	megasr_clist_merge(&linux_adp->build_io_list_in, &clist);

	spin_unlock_irqrestore(linux_adp->host_lock, flags);

	megasr_clist_traverse_each(rclp, rclp_tmp, &clist, rcl_packet_public_t, list) {

		// now try to get the virtual address of the the nodes in the scatter-gather
		// list
		scp		= (struct scsi_cmnd *)rclp->os_packet;
		mlsgl	= (struct scatterlist *)scp->request_buffer;
		sgl		= &rclp->cmn_sgl;

		for (i = 0; i < sgl->sge_count; i++) {
			sgl->sge[i].vaddr	= kmap(mlsgl[i].page) + mlsgl[i].offset;
			if (!sgl->sge[i].vaddr) {
				con_log(CL_ERR, ("megaswr[osl]: kmap failed, it's not supposed to [%d, %d]!\n", sgl->sge_count, i));
#ifdef DEBUG_BUILD
				oss_breakpoint();
				BUG();
#endif
				for (j = 0; j < i; j++) {
					kunmap(mlsgl[j].page);
				}
				// schedule ourselves again, to retry mapping
				spin_lock_irqsave(linux_adp->host_lock, flags);
				megasr_clist_merge(&clist, &linux_adp->build_io_list_in);
				spin_unlock_irqrestore(linux_adp->host_lock, flags);
	
				schedule_work(&linux_adp->work_build_io_in);
				break;
			}
		}

		if (i == sgl->sge_count) {
			// Successfully mapped all pages to virtual address. Remove the RCL packet from the build IO queue
			// and transfer to RCL layer
			megasr_clist_detach(&rclp->list);
			
			spin_lock_irqsave(linux_adp->host_lock, flags);

			osl_cmn_process_rclp(linux_adp->adp, rclp);

			spin_unlock_irqrestore(linux_adp->host_lock, flags);
		}
		else {
			// Do not try to get virtual maps for next commands, we might starve this one.
			break;
		}
	}
}

/**
 * @brief
 * megasr_linux_dpc_complete_cmd
 *
 * counterpart for the "sleepy" build IO routine. We complete the scsi commands to the mid-layer
 * These commands are those for which virtual mapping, using kmap(), was obtained.
 **/
static void megasr_linux_dpc_complete_cmd(void *devp)
{
	linux_adp_t				*linux_adp = (linux_adp_t *)devp;
	rcl_packet_public_t			*rclp;
	rcl_packet_public_t			*rclp_tmp;
	struct scsi_cmnd		*scp;
#ifdef LSI_RAID5_SUPPORT
	big_cmn_sgl_t				*sgl;
#else
	cmn_sgl_t				*sgl;
#endif
	unsigned long			flags;
	struct scatterlist		*mlsgl;
	struct megasr_clist	clist;
	int						i;
	uint8_t					dma_dir;
	

	megasr_clist_setup(&clist);

	spin_lock_irqsave(linux_adp->host_lock, flags);

	megasr_clist_merge(&linux_adp->build_io_list_out, &clist);

	spin_unlock_irqrestore(linux_adp->host_lock, flags);

	megasr_clist_traverse_each(rclp, rclp_tmp, &clist, rcl_packet_public_t, list) {

		scp		= (struct scsi_cmnd *)rclp->os_packet;
		mlsgl	= (struct scatterlist *)scp->request_buffer;
		sgl		= &rclp->cmn_sgl;

		for (i = 0; i < rclp->bkup_sge_count; i++) {
			if (sgl->sge[i].vaddr) {
				kunmap(mlsgl[i].page);
			}
			else {
				con_log(CL_ERR, ("megaswr[osl]: receive request to unmap non-existing mapping\n"));
			}
		}

		megasr_clist_detach(&rclp->list);
			
		spin_lock_irqsave(linux_adp->host_lock, flags);
		
		switch (rclp->io_direction)	{
	
		case IODIR_TO_DEVICE:
			dma_dir	= PCI_DMA_TODEVICE;
			break;

		case IODIR_FROM_DEVICE:
			dma_dir	= PCI_DMA_FROMDEVICE;
			break;

		case IODIR_BIDIRECTIONAL:
		default:
			dma_dir	= PCI_DMA_BIDIRECTIONAL;
			break;
		}

		if (rclp->bkup_sge_count) {		
			if(scp->use_sg) 
				pci_unmap_sg(linux_adp->pdev, (struct scatterlist *)scp->request_buffer, rclp->bkup_sge_count, dma_dir);
			else {
				pci_unmap_single(linux_adp->pdev, (dma_addr_t)((addr_t)sgl->sge[0].paddr), scp->request_bufflen, dma_dir);
				con_log(CL_ERR, ("megasr[osl]: Got called for unmap single in dpc!\n"));
			}
		}

		scp->host_scribble = NULL;
		scp->scsi_done(scp);

		rcl_free_rcl_packet(linux_adp->adp, rclp);
			
		spin_unlock_irqrestore(linux_adp->host_lock, flags);
	}

	spin_lock_irqsave(linux_adp->host_lock, flags);

	/*
	 * Before exiting DPC, check if the work_build_io_out list has any more commands for us to process.
	 * If there are, then reschedule ourselves; reset the dpc_schedule flag otherwise
	 */
	if (!megasr_clist_is_empty(&linux_adp->build_io_list_out)) {
		linux_adp->dpc_scheduled = 1;
		schedule_work(&linux_adp->work_build_io_out);
	}
	else
		linux_adp->dpc_scheduled = 0;

	spin_unlock_irqrestore(linux_adp->host_lock, flags);

}
#endif


/**
 *@brief
 * LSI MegaSR queue command 
 * This function is called by SCSI midlayer when when a new IO command
 * is available for the driver.
 * 
 *@param	scp		scsi command
 *@param	done	call back routine
 *@return	0		If the command successfully passed to the driver
 *@return	1		If the driver busy
 */
int lsraid_queue_cmd(struct scsi_cmnd* scp, void (* done)(struct scsi_cmnd *))
{
	rcl_packet_public_t	*rclp;
	linux_adp_t		*lxadp = (linux_adp_t *)LSI_RAID_SCP2_HOSTDATA(scp);
	scp->scsi_done	= done;

	rclp	= rcl_alloc_rcl_packet(lxadp->adp);
	if (!rclp) {
		return 1;
	}
	rclp->owner = LIN_OSL_RCL_ALLOC1;

	if (scmd_to_rclpkt(scp, rclp, lxadp->pdev) == LSI_TRUE) {
		osl_cmn_process_rclp(lxadp->adp, rclp);
	}
	return 0;
}

/**
 *@brief
 * LSI MegaSWR abort SCSI command. 
 * SCSI midlayer call this function to abort a SCSI command previously sent to
 * the driver. This event happens if the driver fails to finish the command
 * with in SCSI midlayer scsi command time out period.
 * 
 *@param	scp		Pointer to the Scsi Command structure
 *@return	SCSI ABORT status values defined by the SCSI midlayer	 
 *@remarks
 * This is a driver entry function 
 */
int lsraid_abort(struct scsi_cmnd* scp)
{
	int				ret_val = LSI_TRUE;
        linux_adp_t		*linux_adp = ((linux_adp_t *)LSI_RAID_SCP2_HOSTDATA(scp));
	rcl_packet_public_t	*rclp = (rcl_packet_public_t *)scp->host_scribble;
	rcl_packet_public_t	*rclp_tmp;
	rcl_packet_public_t	*rclp_tmp1;

	if (!rclp)
		goto found_null_rclp;

	ret_val = osl_cmn_abort_cmd(linux_adp->adp, rclp);
	
#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,0)
	/* if command aborted successfully, it can be in out bound queue for kunmap process,
	 * Abort function should wait till the command completes kunmap operation.
	 */
	if (ret_val == LSI_TRUE) {
		
		uint8_t	cmd_active_flag;
		
		do {
			cmd_active_flag = 0;
			
			/* find out if this command is present in out bound queue */
			megasr_clist_traverse_each(rclp_tmp, rclp_tmp1, &linux_adp->build_io_list_out, rcl_packet_public_t, list) {
				/* check if command is present in out bound queue */
				if (rclp_tmp == rclp) {
					cmd_active_flag = 1;
					break;
				}
			}	
			
			/* wait if the command is in out bound queue */
			if (cmd_active_flag) {
				spin_unlock_irq(linux_adp->host_lock);
				/* let the system function, dont dead lock */
				msleep(100);
				spin_lock_irq(linux_adp->host_lock);
			}

		} while(cmd_active_flag);
	}
#endif /* LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,0) */

found_null_rclp:

#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,0)
	return ((LSI_TRUE == ret_val) ? SUCCESS: FAILED);
#else
	return ((LSI_TRUE == ret_val) ? SCSI_ABORT_SUCCESS: SCSI_ABORT_SNOOZE);
#endif
}

/**
 *@brief
 * LSI MegaSWR device reset
 * This function is called by the SCSI mid layer if abort command failed.
 * 
 *@param	scp		Pointer to the Scsi Command structure
 *@return	SCSI RESET status values defined by the SCSI midlayer	 
 *@remarks
 * This is a driver entry function 
 */
int lsraid_device_reset(struct scsi_cmnd* scp)
{
	device_id_t	device_id;

	device_id.path		= 0;
	device_id.bus		= LSI_RAID_SCP2_CHANNEL(scp);
	device_id.target	= LSI_RAID_SCP2_TARGET(scp);
	device_id.lun		= LSI_RAID_SCP2_LUN(scp);

	/* Cleanup OS stuff for reset condition and pass the request down to the
	 * next layer */
	if ( LSI_TRUE == osl_cmn_reset(((linux_adp_t *)LSI_RAID_SCP2_HOSTDATA(scp))->adp, DEV_RESET, &device_id) )
		return SUCCESS;
	else
		return FAILED;
}

/**
 *@brief
 * LSI MegaSWR bus reset
 * This function is called by the SCSI mid layer if reset device failed.
 *
 *@param	scp		Pointer to the Scsi Command structure
 *@return	SCSI RESET status values defined by the SCSI midlayer	 
 *@remarks
 * This is a driver entry function 
 */
int lsraid_bus_reset(struct scsi_cmnd* scp)
{
	device_id_t	device_id;

	device_id.path		= 0;
	device_id.bus		= LSI_RAID_SCP2_CHANNEL(scp);
	device_id.target	= LSI_RAID_SCP2_TARGET(scp);
	device_id.lun		= LSI_RAID_SCP2_LUN(scp);

	/* Cleanup OS stuff for reset condition and pass the request down to the
	 * next layer */
	if ( LSI_TRUE == osl_cmn_reset(((linux_adp_t *)LSI_RAID_SCP2_HOSTDATA(scp))->adp, BUS_RESET, &device_id) )
		return SUCCESS;
	else
		return FAILED;
}

/**
 *@brief
 * LSI MegaSWR host reset
 * This function is called by the SCSI mid layer if reset bus failed.
 *
 *@param	scp		Pointer to the Scsi Command structure
 *@return	SCSI RESET status values defined by the SCSI midlayer	 
 *@remarks
 * This is a driver entry function 
 */
int lsraid_host_reset(struct scsi_cmnd* scp)
{
	device_id_t	device_id;

	device_id.path		= 0;
	device_id.bus		= LSI_RAID_SCP2_CHANNEL(scp);
	device_id.target	= LSI_RAID_SCP2_TARGET(scp);
	device_id.lun		= LSI_RAID_SCP2_LUN(scp);

	/* Cleanup OS stuff for reset condition and pass the request down to the
	 * next layer */
	if ( LSI_TRUE == osl_cmn_reset(((linux_adp_t *)LSI_RAID_SCP2_HOSTDATA(scp))->adp, ADP_RESET, &device_id) )
		return SUCCESS;
	else
		return FAILED;
}

/**
 *@brief
 * LSI MegaSWR Interrupt Service Routine 
 * This function is called by the SCSI Midlayer if an ISR event happen.
 * 
 *@param irq	IRQ number
 *@param os_cxt	: Driver extension data structure pointer
 *@param regs	: pt_regs structure pointer 
 *@return None
 *@remarks
 * This is a driver entry function 
 */
static LSI_RAID_IRQRETURN_T lsraid_isr(int irq, void* os_cxt, 
		struct pt_regs* regs)
{
	int				ret_val = LSI_FALSE;
	unsigned long	flags;
	linux_adp_t		*lx_adp	= os_cxt;

	spin_lock_irqsave(lx_adp->host_lock, flags);

	ret_val = megasr_lib_isr(lx_adp->adp);

	/* @todo should be done from here and needs to be removed from success failure handler */
	osl_cmn_flush_schedule_queue(lx_adp->adp);

	spin_unlock_irqrestore(lx_adp->host_lock, flags);

	if (ret_val == LSI_TRUE) {
		return IRQ_RETVAL(1);
	} else {
		return IRQ_RETVAL(0);
	}
}

/**
 *@brief
 * LSI megaswr Initialize adapter 
 * Perform complete initialization for one adapter.
 *
 *@param	pdev	linux specific pci device structure
 *@param	id		Set of pci attributes of this adapter
 *@return	0		If success
 *@return	Non zero scsi error value in case of failure
 *@remarks
 * This function is used by both 2.6 and 2.4 kernel drivers. 
 */
static int __devinit lsraid_probe_one(struct pci_dev* pdev,
		const struct pci_device_id *id)
{
	pvt_data_t			*adp	= NULL;
	linux_adp_t			*lxadp		= NULL;
	struct Scsi_Host	*host		= NULL;
	uint32_t			i			= 0;

	for (i = 0; i < adp_count_g ; i++) {
	
		lxadp = (linux_adp_t*)lin_get_linux_adp(adps_g[i]); 
		
		if ((lxadp->pdev->bus->number == pdev->bus->number) && 
				(lxadp->pdev->devfn == pdev->devfn)) {
			con_log(CL_ERR,("megaswr[osl]:  reject probe request for dev\n"));

			con_log(CL_ERR,("megaswr[osl]: %#4.04x:%#4.04x:%#4.04x:%#4.04x:",
				pdev->vendor, pdev->device, pdev->subsystem_vendor,
				pdev->subsystem_device));

			return -ENODEV;
		}
	}

	// Announce PCI information for the probed controller
	con_log(CL_ANN, ("megasr: [RAID HBA] %#4.04x:%#4.04x:%#4.04x:%#4.04x: bus %d:slot %d:func %d\n",
		pdev->vendor, pdev->device, pdev->subsystem_vendor, pdev->subsystem_device,
		pdev->bus->number, PCI_SLOT(pdev->devfn), PCI_FUNC(pdev->devfn)));


	if(pci_enable_device(pdev))
		return -ENODEV;

	// Get the size of adapter structure
	// allocate memory, zero out
	// link Linux adapter and adp
	adp = kmalloc(megasr_lib_get_adp_size(), GFP_KERNEL);
	if ( !adp )
		return -ENODEV;
	
	memset(adp, 0, megasr_lib_get_adp_size());

	lxadp = kmalloc(sizeof(linux_adp_t), GFP_KERNEL);
	if (!lxadp){
		goto lsraid_fail_step1;
	}

	memset(lxadp, 0, sizeof(linux_adp_t));

	lxadp->adp = adp;
	megasr_lib_set_lxadp(adp, lxadp);

	lxadp->pdev	= pdev;

	/* Store pci config info in adp */
	extract_pciconfig(pdev, adp);

	if( LSI_FALSE == 
			megasr_init_adp(adp, lxadp->pdev->bus->number, 
					(lxadp->pdev->devfn >> 3), (lxadp->pdev->devfn & 7))) {
		con_log(CL_WARN,("megaswr[osl]:  rcl_claim_adapter failed\n"));
		goto lsraid_fail_step2;
	}

	if (allocate_static_mem_and_add_to_memmgr(lxadp) != LSI_TRUE) {
		con_log(CL_WARN,("megaswr[osl]:  allocate mem and add to memmgr failed!!\n"));
		goto lsraid_fail_step2;
	}

	if(rcl_init_adapter(adp) == LSI_FALSE) {
		con_log(CL_WARN,("megaswr[osl]:  rcl_init_adapter failed\n"));
		goto lsraid_fail_step2;
	}

#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,0)
	// setup work queues for DPC
	megasr_clist_setup(&lxadp->build_io_list_in);
	megasr_clist_setup(&lxadp->build_io_list_out);
	INIT_WORK(&lxadp->work_build_io_in, megasr_linux_build_io, lxadp);
	INIT_WORK(&lxadp->work_build_io_out, megasr_linux_dpc_complete_cmd,
	lxadp);
#endif

#ifdef __LP64__
	if (pci_set_dma_mask(pdev, 0xFFFFFFFFFFFFFFFFULL) != 0) {
		con_log(CL_ERR, ("megaswr[osl]: failed to set the DMA mask\n"));

		if (pci_set_dma_mask(pdev, 0xFFFFFFFF) != 0) {
			con_log(CL_ERR, ("megaswr[osl]: failed to set the DMA mask again\n"));
			goto lsraid_fail_step3;
		}
	}
	else {
		con_log(CL_INFO, ("megaswr[osl]: FLAG FFFFFFFFFFFFFFFF set\n"));	
	}
#else
	if (pci_set_dma_mask(pdev, 0xFFFFFFFF) != 0) {
		con_log(CL_ERR, ("megaswr[osl]: failed to set the DMA mask\n"));
		goto lsraid_fail_step3;
	}
#endif
		
	host = LSI_RAID_SCSI_REGISTER(lsraid_template_gp, sizeof(linux_adp_t *));
	if(!host) {
		con_log(CL_ERR, ("megaswr[osl]:  scsi_register failed\n"));
		goto lsraid_fail_step3;
	}

	/* to make sure the SCSI Mid Layer IOCTL Works */
	LSI_RAID_SCSI_SET_PCI_DEVICE(host, pdev);

	/* Init Global Lock */
	spin_lock_init(&lxadp->lock);

	lxadp->host_lock	= &lxadp->lock;

	/* To handle kernels with out host_lock */
	LSI_RAID_HOST_SET_LOCK(host, lxadp->host_lock);
	/* to make sure the SCSI Mid Layer IOCTL Works */
	LSI_RAID_SCSI_SET_PCI_DEVICE(host, pdev);

	host->hostdata[0] = (unsigned long) lxadp;

	lxadp->host		= host;

	host->this_id		= -1;
	host->unique_id		= pdev->bus->number << 8 | pdev->devfn;
	host->max_id		= MAX_LD_PER_ADP;
	host->max_lun		= 1;
	host->max_channel	= LD_BUS;
	host->base			= pci_resource_start(pdev, 1);
	host->irq			= pdev->irq;
	host->max_cmd_len	= MAX_CDB_LENGTH;
	megasr_lib_init_adp(adp, &host->cmd_per_lun, &host->max_sectors, adp_count_g);
	host->can_queue		= MAX_IO_PER_ADP;
	host->sg_tablesize	= MAX_SGE_COUNT_FOR_OS;
	
	if (rcl_scan_devices(adp) == LSI_FALSE) {
		con_log(CL_ERR,("megaswr[osl]:  rcl scan devices failed\n"));
		goto lsraid_fail_init;
	}
		
	/* setup adapter handle in PCI soft state */
	pci_set_drvdata(pdev, host);

	/* set up ioctl interface for this adapter */
	if (lsraid_load_ioctl_module(adp) == LSI_FALSE) {
		con_log(CL_ERR,("megaswr[osl]:  load ioctl module failed\n"));
		goto lsraid_fail_init;
	}

	if(request_irq(pdev->irq, lsraid_isr, SA_SHIRQ, "megasr", lxadp)) {
		con_log(CL_ERR,("megaswr[osl]:  request_irq failed\n"));
		goto lsraid_fail_init;
	}

	/* configuration load function call */
	if (rcl_read_and_populate_cfg(adp) == LSI_FALSE) {
		con_log(CL_ERR,("megaswr[osl]:  configuration resolve failed\n"));
		goto lsraid_fail_init;
	}

	/* raid key authentication call */
	rcl_authenticate_ibutton(adp);

	if (LSI_RAID_SCSI_ADD_HOST(host, &pdev->dev)) {
		con_log(CL_ERR,("megaswr[osl]:  scsi add host failed\n"));
		goto lsraid_fail_init;
	}

	LSI_RAID_SCSI_SCAN_HOST(host);
	
	adp_count_g++;

#if LINUX_VERSION_CODE < KERNEL_VERSION(2,6,0)
	if(megasas_proc_hba_map)
		megasas_create_proc_entry(lxadp);
#endif
	return 0;

lsraid_fail_init:
	LSI_RAID_SCSI_HOST_DEALLOC(host);

lsraid_fail_step3:
	rcl_unload(adp);

	osl_destroy_all_timers(lxadp);

lsraid_fail_step2:
	kfree((void *)lxadp);

lsraid_fail_step1:
	kfree(adp);

	return -ENODEV;
}

/**
 *@brief
 * LSI megaswr Detach one adapter 
 * Perform complete unloading of driver for one adapter
 *
 *@param	dev		linux specific pci device structure
 *@return	None 
 *@remarks
 * This function is used by both 2.6 and 2.4 kernel drivers. 
 */
static void lsraid_detach_one(struct pci_dev *dev)
{
	struct Scsi_Host	*host = pci_get_drvdata(dev);
	linux_adp_t			*lxadp	= (linux_adp_t *)host->hostdata[0];
    unsigned long       flags;

	con_log(CL_OSL, ("megaswr[osl]: driver detach function called\n"));

	/* shut down all active operations, and prepare for system shut down */
    spin_lock_irqsave(lxadp->host_lock, flags);
	rcl_shutdown(lxadp->adp);
    spin_unlock_irqrestore(lxadp->host_lock, flags);
	
#if LINUX_VERSION_CODE < KERNEL_VERSION(2,6,0)
	
	if (megasas_proc_root && lxadp->hba_proc_dir) {
		megasas_remove_proc_entry(lxadp);
	}

	if (adp_count_g == 1) {
		if (megasas_proc_root) {
			remove_proc_entry("version", megasas_proc_root);
			remove_proc_entry("release_date", megasas_proc_root);
			remove_proc_entry("hba_map", megasas_proc_root);
			remove_proc_entry(LSIRAID_DRIVER_SIGNATURE, proc_scsi);
		}
	}
#endif
	con_log(CL_DEBUG, ("megaswr[osl]:  Releasing the controller... "));

	osl_destroy_all_timers(lxadp);

	LSI_RAID_SCSI_REMOVE_HOST(host);

	free_irq(lxadp->pdev->irq, lxadp);

	rcl_unload(lxadp->adp);

	remove_mem_allocated_for_memmgr(lxadp);
	
	if (adp_count_g == 1) {
		lsraid_unload_ioctl_module(lxadp);
	}

	kfree(lxadp->adp);

	kfree(lxadp);

	LSI_RAID_SCSI_HOST_DEALLOC(host);
	
	adp_count_g--;
}

/**
 *@brief
 * Extract PCI configuration data
 * This function is used to extract PCI configuration data from pci_dev
 * structure to driver provided structure pci_config_space.
 * 
 *@param pci_dev : linux specific pci device structure
 *@param pci_cfg : swraid specific pci device structure
 *@return None
 */
static void extract_pciconfig(struct pci_dev* pdev, pvt_data_t* adp)
{
	int			i;
	uint32_t	cfg_size = 0;
	uint8_t		*ptr;

	ptr = lin_get_adp_pci_config_space_ptr_and_size(adp, &cfg_size);

	for(i = 0; i < cfg_size; ptr++, i++)
		pci_read_config_byte(pdev, i, ptr);

	for(i = 0; i < MAX_PCI_ELEMENTS; ptr++, i++)
		lin_set_adp_pci_config(pci_resource_start(pdev,i),i, adp);

	return;
}

/**
 *@brief
 * This is used to de-allocate the memory allocated for memory manager.
 */
void remove_mem_allocated_for_memmgr(linux_adp_t *lxadp)
{
	while(lxadp->dma.list_count) {
		lxadp->dma.list_count--;
		megasr_dma_free(lxadp->pdev, &lxadp->dma.list[lxadp->dma.list_count]);
	}
}

/**
 *@brief
 * This is used to allocate the required memory for linux driver and set up
 * the memory manager. Adjustments of required memory is done inside this
 * function. Fail condition of this function is fully implemented. 
 */
boolean_t allocate_static_mem_and_add_to_memmgr(linux_adp_t *lxadp)
{
	pvt_data_t *adp		= lxadp->adp;
	linux_dma_mem_t		*dma = &(lxadp->dma);
	uint32_t 			try_size;
	/* static dma pool size, windows driver allocates TOTAL_STATIC_MEM_REQ	*/
	uint32_t			static_pool_len = LSI_RAID_1_MB_SIZE/2;
	uint32_t			static_dma_pool_len = 2 * LSI_RAID_1_MB_SIZE;
	cmn_sgl_t			dyn_sgl;
#if LINUX_VERSION_CODE < KERNEL_VERSION(2,6,0)
	uint8_t				alloc_flag = LINUX_DMA_MEM_GFP;
#else
	uint8_t				alloc_flag = LINUX_DMA_NOWAIT;
#endif

#ifdef LSI_RAID5_SUPPORT
	mem_mgr_init_mem_mgr(adp,(64*1024)); // Initialize with 64k node size
#else
	mem_mgr_init_mem_mgr(adp, (32*1024)); // Initialize with 32k node size
#endif

	lxadp->static_mem = NULL;
	dma->list_count = 0;
	
	if (megasr_dma_alloc(lxadp->pdev, LSI_RAID_1_MB_SIZE, &dma->list[dma->list_count], LINUX_DMA_NOWAIT) != LSI_TRUE) {
		con_log(CL_ERR, ("megasr[osl]:  out of memory in %s for VA only pool\n", __FUNCTION__));
		goto lsraid_fail_step1;
	}
	
	if (LSI_TRUE != mem_mgr_add_to_mem_pool(adp, dma->list[dma->list_count].vir,	NULL, dma->list[dma->list_count].len, NO_PHY_ADDR_MEM_POOL)) {
		con_log(CL_ERR, ("megasr[osl]:  add static memory to the memory pool failed\n"));
		goto lsraid_fail_step1;
	}

	dma->list_count++;

	try_size = static_pool_len;
	while (static_pool_len) {

		if (dma->list_count >= MAX_MEM_POOLS_PER_POOL_TYPE) {
			con_log(CL_ERR, ("megasr[osl]:  memory allocation failed, overflow\n"));
			goto lsraid_fail_step1;
		}

		if (megasr_dma_alloc(lxadp->pdev, try_size, &dma->list[dma->list_count], LINUX_DMA_NOWAIT) != LSI_TRUE) {
			try_size /= 2;
			if (try_size < (LSI_RAID_1_MB_SIZE / 8)) {
				/* i cannot try any more and giving up */
				con_log(CL_ERR, ("megasr[osl]:  warning, driver allocated %#x bytes less than requested memory\n", static_pool_len));
				return LSI_TRUE;
			}
			continue;
		}

		if (LSI_TRUE != mem_mgr_add_to_mem_pool(adp, dma->list[dma->list_count].vir, (uint8_t *)(addr_t)dma->list[dma->list_count].dma_addr, try_size, STATIC_MEM_POOL)) {
				con_log(CL_ERR, ("megasr[osl]:  add memory to the memory pool failed\n"));
				goto lsraid_fail_step1;
		}
	
		static_pool_len -= try_size; 
		dma->list_count++;
	}
	
	// add memory for initial dynamic buffers
	try_size = static_dma_pool_len;
	while (static_dma_pool_len) {

		if (dma->list_count >= MAX_MEM_POOLS_PER_POOL_TYPE) {
			con_log(CL_ERR, ("megasr[osl]:  memory allocation failed, overflow\n"));
			goto lsraid_fail_step1;
		}

		if (megasr_dma_alloc(lxadp->pdev, try_size,	&dma->list[dma->list_count], alloc_flag) != LSI_TRUE) {
			try_size /= 2;
			if (try_size < (LSI_RAID_1_MB_SIZE / 8)) {
				/* i cannot try any more and giving up */
				con_log(CL_ERR, ("megasr[osl]:  warning, driver allocated %#x bytes less than requested dynamic memory\n", static_dma_pool_len));
				return LSI_TRUE;
			}
			continue;
		}

		oss_memset((uint8_t *)&dyn_sgl, sizeof(cmn_sgl_t), 0);

		// call the fn to set the desc size in mem mgr
		mem_mgr_get_desc_size_for_sgl_mem(adp, try_size);

		dyn_sgl.sge_count = 1;
		dyn_sgl.sge[0].length = try_size;
		dyn_sgl.sge[0].vaddr = dma->list[dma->list_count].vir;
		dyn_sgl.sge[0].paddr = (uint8_t *)(unsigned long)dma->list[dma->list_count].dma_addr;	// This is broken till we are capable of supported dma >4GB on 32-bit plaforms :-(
		if (mem_mgr_add_sgl_mem(adp, &dyn_sgl, NULL) == LSI_FALSE) {
			con_log(CL_ERR, ("megasr[osl]: can't add mem to sgl pool for static dma\n"));
			goto lsraid_fail_step1;
		}

		static_dma_pool_len -= try_size; 
		dma->list_count++;
	}

	return LSI_TRUE;

lsraid_fail_step1:
	remove_mem_allocated_for_memmgr(lxadp);

	return LSI_FALSE;
}

/**
 * @brief
 * Convert SCSI command to RCL packet
 * This is used to convert scsi cmd to rcl packet command.  We detect if the IO is
 * meant for a RAID 5 volume and if it is a READ/WRITE command. In this case, the
 * tasklet to map the virtual address is scheduled. Otherwise, send the command
 * directly to RCL for execution.
 * 
 * @param scp	scsi mid layer command
 * @param rclp	rcl packet
 * @param dev	linux specific pci device
 *
 * @return LSI_TRUE if this command can be issue right away
 * @return LSI_FALSE if this command is scheduled for tasklet
 **/
static uint32_t scmd_to_rclpkt(struct scsi_cmnd *scp, rcl_packet_public_t *rclp, struct pci_dev *dev)
{
	linux_adp_t		*linux_adp;
	int				i;
#ifdef LSI_RAID5_SUPPORT
	big_cmn_sgl_t		*sgl;
#else
	cmn_sgl_t		*sgl;
#endif
	int				sgcnt;
	dma_addr_t		dmah;
	struct scatterlist	*mlsgl;
	uint8_t			dma_dir;
	scsi_packet_t	*sp;
	uint32_t		rc = LSI_TRUE;

	linux_adp	= (linux_adp_t *)LSI_RAID_SCP2_HOSTDATA(scp);

	scp->host_scribble		= (uint8_t*) rclp;
	rclp->os_packet			= (pvt_data_t *)scp;
	rclp->device_id.path	= 0;
	rclp->device_id.bus		= LSI_RAID_SCP2_CHANNEL(scp);
	rclp->device_id.target	= LSI_RAID_SCP2_TARGET(scp);
	rclp->device_id.lun		= LSI_RAID_SCP2_LUN(scp);
	rclp->cmd.packet_type	= SCSI_PACKET;
	rclp->buffer_len		= scp->request_bufflen;
	/* both success and failure hanlers are same */
	rclp->success_handler	= (rclp_callback_t)osl_cmd_pkt_success_handler;
	rclp->failure_handler	= (rclp_callback_t)osl_cmd_pkt_success_handler;

#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,0)
	if(scp->sc_data_direction == DMA_TO_DEVICE) {
		dma_dir				= PCI_DMA_TODEVICE;
		rclp->io_direction	= IODIR_TO_DEVICE;
	} else if(scp->sc_data_direction == DMA_FROM_DEVICE) {
		dma_dir				= PCI_DMA_FROMDEVICE;
		rclp->io_direction	= IODIR_FROM_DEVICE;
	} else if(scp->sc_data_direction == DMA_NONE) {
		dma_dir				= PCI_DMA_NONE;
		rclp->io_direction	= IODIR_NONE;
	} else {
		dma_dir				= PCI_DMA_BIDIRECTIONAL;
		rclp->io_direction	= IODIR_BIDIRECTIONAL;
	}
#else
	if(scp->sc_data_direction == SCSI_DATA_WRITE) {
		dma_dir				= PCI_DMA_TODEVICE;
		rclp->io_direction	= IODIR_TO_DEVICE;
	} else if(scp->sc_data_direction == SCSI_DATA_READ) {
		dma_dir				= PCI_DMA_FROMDEVICE;
		rclp->io_direction	= IODIR_FROM_DEVICE;
	} else if(scp->sc_data_direction == SCSI_DATA_NONE) {
        dma_dir             = PCI_DMA_NONE;
        rclp->io_direction  = IODIR_NONE;
	} else {
		dma_dir				= PCI_DMA_BIDIRECTIONAL;
		rclp->io_direction	= IODIR_BIDIRECTIONAL;
	}
#endif

	sgl		= &rclp->cmn_sgl;
	sp		= &rclp->cmd.u.scsi_packet;
	sp->cdb_len	= scp->cmd_len;

	memcpy(sp->u.lsi_cdb, scp->cmnd, scp->cmd_len);
	
	switch (sp->cdb_len) {
	case 6:
	case 10 :
	case 12:
		rclp->start_blk = GET_START_BLOCK(sp->u.lsi_cdb); 
		rclp->blk_count = GET_NUMBER_OF_BLOCKS(sp->u.lsi_cdb);
		break;
	case 16:
		rclp->start_blk = GET_START_BLOCK16(sp->u.lsi_cdb); 
		rclp->blk_count = GET_NUMBER_OF_BLOCKS16(sp->u.lsi_cdb);
		break;
	}

	rclp->status = STS_SUCCESS;
	lin_set_rclp_ld_ptr(linux_adp->adp, rclp, &rclp->device_id);

	if (!scp->use_sg) {
		if (scp->request_bufflen) {
			dmah = pci_map_single(dev, scp->request_buffer, scp->request_bufflen, dma_dir);

			sgl->sge_count		= 1;
			sgl->sge[0].vaddr	= scp->request_buffer;
			sgl->sge[0].paddr	= (void *) (addr_t)dmah;
			sgl->sge[0].length	= scp->request_bufflen;
		}
		else {
			sgl->sge_count = 0;
		}

		rclp->bkup_sge_count = sgl->sge_count;
		return LSI_TRUE;
	}

	mlsgl = (struct scatterlist *)scp->request_buffer;
	
	sgcnt = pci_map_sg(dev, mlsgl, scp->use_sg, dma_dir);

#ifdef DEBUG_BUILD
	
	/* This is to check whether os is sending i/o with size more than the requested i/o size and
	 * with more than the requested sge count for a single sgl
	*/

#ifdef ENABLE_COALESCING
	BUG_ON(scp->request_bufflen > COALESCING_MAX_SIZE_R5);
#endif
	BUG_ON(sgcnt > MAX_SGE_COUNT_FOR_OS);
#endif //DEBUG_BUILD

	rclp->buffer_len = 0;

	for (i = 0; i < sgcnt; i++) {
		sgl->sge[i].paddr	= (void *)(addr_t)sg_dma_address(&mlsgl[i]);
		sgl->sge[i].length	= sg_dma_len(&mlsgl[i]);
		rclp->buffer_len	+= sgl->sge[i].length;
#if LINUX_VERSION_CODE < KERNEL_VERSION(2,6,0)
		sgl->sge[i].vaddr  = mlsgl[i].address;
#endif
	}

	sgl->sge_count = sgcnt;

#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,0)
	// If this is a RAID 5 IO, or a non-IO command, we would require virtual address
	// of the data transfer buffer. Return LSI_FALSE to indicate this.
	if ( rcl_is_r5rw_or_noniocmd(linux_adp->adp, rclp)){
		megasr_clist_append(&rclp->list, &linux_adp->build_io_list_in);
		rc = LSI_FALSE;
		schedule_work(&linux_adp->work_build_io_in);
	}
#endif
	rclp->bkup_sge_count = sgl->sge_count;
	return rc;
}

static void_t osl_destroy_all_timers(linux_adp_t *lxadp)
{
	uint8_t		i;

	for (i= 0; i < MAX_TIMERS; i++) {
		if(lxadp->timer_objects[i].active_flag) {
			del_timer(&(lxadp->timer_objects[i].timer));
			lxadp->timer_objects[i].active_flag	= 0;
			lxadp->timer_count --;
			con_log(CL_ERR, ("megaswr[osl]:  err timer[%d] are not cleared\n", i));
		}
	
		lxadp->timer_objects[i].used_flag	= 0;
	}
}

/**
 *@brief
 * Timer timeout function call to terminate the timer scheduled by the other
 * layers
 * 
 *@param	data	: Pointer to timer object structure. 
 *@return	None 
 */
void osl_timeout(unsigned long data)
{
	timer_object_t*	to = (timer_object_t*) data;
	to->active_flag = 0;
	to->timer_fn(to->context1, to->context2); 
}

/**
 *@brief
 * Function used to allocate pools of memory to satisfy RAID5 memory requirements when
 * a RAID5 LD is detected or created. The caller should not expect this memory to be
 * available immediately, hence the reason the callback exists. The caller's callback
 * will be invoked once the memory has been allocated, virtual addresses have been mapped,
 * and the SGL has been added to the memory manager, ready for use. Please note that the
 * caller is not guaranteed to get the total amount of memory requested; when this function returns
 * to the caller, the size allocated will be passed back, informing that subsystem
 * about how much memory the OS had available at that time. The caller then can retry for any
 * remaining amount at a later time if they choose to. The callback is invoked after the memory
 * that could be allocated is added to the memory manager and is ready for use.
 * 
 *@param	os_context		: Pointer to os context data.
 *@param	size_in_bytes	: Amount of memory in bytes to allocate and add to memory manager.
 *@param size_allocated	: OS may not give us the requested amount of memory, the amount we are able to get will be returned in this parameter.
 *@param	callback		: Callback routine to notify when memory is available.
 *@param callback_context: Callback routine's context data.
 *
 *@return LSI_TRUE:	A portion or all of the requested memory has been allocated and the callback
 *					will be invoked when memory is ready for use. Caller should not assume that
 *					the memory is available and/or that all of the memory could be allocated
 *					simply because LSI_TRUE was returned.
 *		LSI_FALSE:	No memory could be allocated, not even a portion of the requested size.
 */
boolean_t osl_allocate_pool(pvt_data_t *os_context, uint32_t size_in_bytes, uint32_t *size_allocated, mem_added_fn_t callback, pvt_data_t *callback_context)
{
	linux_adp_t			*lxadp	= (linux_adp_t *)lin_get_linux_adp(os_context);
	linux_dma_mem_t		*dma = &lxadp->dma;
	cmn_sgl_t			dyn_sgl;
	uint32_t			remaining_allocation = size_in_bytes;
	uint32_t			try_size;
#if LINUX_VERSION_CODE < KERNEL_VERSION(2,6,0)
	uint8_t				alloc_flag = LINUX_DMA_MEM_GFP;
#else
	uint8_t				alloc_flag = LINUX_DMA_NOWAIT;
#endif

	/* Validate data passed in */
	if (!size_in_bytes) {
		con_log(CL_ERR, ("megaswr[osl]:  err allocate_pool called with zero size to allocate!\n"));
		return LSI_FALSE;
	}
	if (!size_allocated) {
		con_log(CL_ERR, ("megaswr[osl]: err allocate_pool called with null pointer to return size allocated!\n"));
		return LSI_FALSE;
	}
	if (!callback) {
		con_log(CL_ERR, ("megaswr[osl]:  err allocate_pool called without a callback!\n"));
		return LSI_FALSE;
	}

	try_size = LSI_RAID_1_MB_SIZE;
	while (remaining_allocation) {

		if (try_size > remaining_allocation) {
			try_size = remaining_allocation;
		}

		if (dma->list_count >= MAX_MEM_POOLS_PER_POOL_TYPE) {
			con_log(CL_ERR, ("megasr[osl]:  memory allocation failed, overflow\n"));
			return LSI_FALSE;
		}

		if (megasr_dma_alloc(lxadp->pdev, try_size, &dma->list[dma->list_count],  alloc_flag) != LSI_TRUE) {
			try_size /= 2;
			if (try_size < (LSI_RAID_1_MB_SIZE / 8)) {
				/* i cannot try any more and giving up */
				con_log(CL_DEBUG, ("megasr[osl]:  warning, driver allocated %#x bytes less than requested memory for dynamic\n", size_in_bytes));
				break;
			}
			continue;
		}
		
		dma->list[dma->list_count].len = try_size;

		oss_memset((uint8_t *)&dyn_sgl, sizeof(cmn_sgl_t), 0);

		dyn_sgl.sge_count = 1;
		dyn_sgl.sge[0].length = try_size;
		dyn_sgl.sge[0].vaddr = dma->list[dma->list_count].vir;
		dyn_sgl.sge[0].paddr = (uint8_t *)(unsigned long)dma->list[dma->list_count].dma_addr;	// This is broken till we are capable of supported dma >4GB on 32-bit plaforms :-(
		
		if (mem_mgr_add_sgl_mem(lxadp->adp, &dyn_sgl, callback_context) == LSI_FALSE) {

			con_log(CL_ERR, ("megasr[osl]: can't add mem to sgl pool for dynamic buffers [%#x:%#x]\n",
					size_in_bytes, remaining_allocation));

			megasr_dma_free(lxadp->pdev, &lxadp->dma.list[lxadp->dma.list_count]);
			break;
		}

		remaining_allocation	-= try_size; 
		try_size				= LSI_RAID_1_MB_SIZE;
		dma->list_count++;
	}

	*size_allocated = size_in_bytes - remaining_allocation;

	if (*size_allocated) {
		callback(os_context, callback_context);
		return LSI_TRUE;
	}
	else {
		return LSI_FALSE;
	}
}

/**
 *@brief 
 * Complete the given rcl packet to the OS With SUCCESS state.
 *
 *@param	os_cxt	Complete adapter soft-state
 *@param	rclp	RCL packet pointer
 *@return	LSI_TRUE always	
 */
void_t osl_cmd_pkt_success_handler(pvt_data_t* os_cxt, pvt_data_t* rclp_ptr)
{
	rcl_packet_public_t		*rclp = (rcl_packet_public_t *)rclp_ptr;
	struct scsi_cmnd	*scp = (struct scsi_cmnd *)rclp->os_packet;
	linux_adp_t 		*lxadp ;
	uint8_t				dma_dir;
	uint8_t				complete_in_dpc;
#ifdef LSI_RAID5_SUPPORT
	big_cmn_sgl_t			*sgl;
#else
	cmn_sgl_t			*sgl;
#endif
#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,0)
	scsi_packet_t		*sp;
#endif

	lxadp = (linux_adp_t*)lin_get_linux_adp(os_cxt); // at this point Lin OSL do not have Linux adapter and so we need to use this call
	switch (rclp->io_direction)	{
	
		case IODIR_TO_DEVICE:
			dma_dir	= PCI_DMA_TODEVICE;
			break;

		case IODIR_FROM_DEVICE:
			dma_dir	= PCI_DMA_FROMDEVICE;
			break;

		case IODIR_BIDIRECTIONAL:
		default:
			dma_dir	= PCI_DMA_BIDIRECTIONAL;
			break;
	}


	if (rclp->status == STS_SUCCESS) {
		scp->result	= 0;
	}
	else {
		switch(rclp->status) {
		case STS_UNSUPPORTED:	// FALL THROUGH
		case STS_NO_DEVICE:
			scp->result = (DID_BAD_TARGET << 16); 
			break;

		/* handle media error properly */
		case STS_MEDERR:
			osl_cmn_fill_sense_info_from_rclp(rclp_ptr, (void_t*)scp->sense_buffer, SCSI_SENSE_BUFFERSIZE);
			scp->result = (DRIVER_SENSE << 24) | (DID_ERROR << 16) | (CHECK_CONDITION << 1);
			break;

		case STS_CD_DVD_ERROR:
			scp->result = (DRIVER_SENSE << 24) | (DID_OK << 16) | (CHECK_CONDITION << 1);
			break;

		case STS_CLEAN_ABORT:	// FALL THROUGH
		case STS_DIRTY_ABORT:	// FALL THROUGH
		case STS_TIME_OUT:		// FALL THROUGH
		case STS_ERROR:			// FALL THROUGH
		case STS_RESET:			// FALL THROUGH
		default:
			scp->result = (DID_ERROR << 16);
			break;
		}

		scp->result |= rclp->scsi_status;

        if (rclp->device_id.bus < LD_BUS && rclp->sense_len)
            oss_memcpy((uint8_t*)&rclp->sense_info, scp->sense_buffer, sizeof(lsi_sense_data));
	}

	complete_in_dpc = 0;

#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,0)
	// If this packet requires kunmap, schedule DPC for it
	sp = &rclp->cmd.u.scsi_packet;
	if (scp->use_sg && (rcl_is_r5rw_or_noniocmd(lxadp->adp, rclp)) ) {
		megasr_clist_append(&rclp->list, &lxadp->build_io_list_out);
		complete_in_dpc = 1;

		/* Don't reschedule the dpc if it is already done; but if this is the last command we are ever going
		 * to get, then the DPC may not process this. To avoid this, we will check if this build_io_list_out
		 * once more before exiting DPC
		 */
		if (!lxadp->dpc_scheduled) {
			lxadp->dpc_scheduled = 1;
			schedule_work(&lxadp->work_build_io_out);
		}
	}
#endif

	if (!complete_in_dpc) {

		/* Check if pci mapping is done */
		if (rclp->bkup_sge_count) {		
			sgl	= &rclp->cmn_sgl;
			if(scp->use_sg) {
				pci_unmap_sg(lxadp->pdev, (struct scatterlist *)scp->request_buffer, rclp->bkup_sge_count, dma_dir);
			}
			else {
				pci_unmap_single(lxadp->pdev, (dma_addr_t)((addr_t)sgl->sge[0].paddr), scp->request_bufflen, dma_dir);
			}
		}
		
		scp->host_scribble = NULL;
		scp->scsi_done(scp);

		rcl_free_rcl_packet(os_cxt, rclp);
	}

	/* flush commands in schedule queue */
	osl_cmn_flush_schedule_queue(os_cxt);
}
uint32_t lin_get_host_no(linux_adp_t * lxadp)
{
	return lxadp->host->host_no;
}

int
megasr_printk(const char *fmt, ...)
{
	va_list		ap;
	char		buf[256];

	va_start(ap, fmt);
	vsprintf(buf, fmt, ap);
	va_end(ap);
	return printk(buf);
}

module_init(lsraid_init);
module_exit(lsraid_exit);

/* vim: set ts=4 sw=4 tw=78 nowrap ai si: */
